//
// Hi-Tech Lite source file for DB038 board with 16F887:
// LCD functions
//
// $Id: db038-lcd.c,v 1.3 2012/04/24 17:25:23 Wouter Exp $
//
// (c) Wouter van Ooijen - wouter@voti.nl
//
// Copying and distribution of this file, with or without modification,
// are permitted in any medium without royalty provided the copyright
// notice and this notice are preserved.  This file is offered as-is,
// without any warranty.
// 

#include "db038.h"
#include "db038-lcd.h"

#define LCD_E( x ) do { PIN_COPY( PORTA, 5, x ); } while(0)
void lcd_e( unsigned char x ){
   LCD_E( x );
}

#define LCD_RS( x ) do { PIN_COPY( PORTD, 3, x); } while(0)
void lcd_rs( unsigned char x ){
   LCD_RS( x );
}

#ifdef DB038_PIC14
  
   #define LCD_DATA_4( x ){ \
      PORTD_SHADOW &= 0x0F; \
      PORTD_SHADOW |= (( x ) << 4 ); \
      PORTD_FLUSH;	\
      WAIT_US( 20 ); \
      LCD_E( 1 ); \
      WAIT_US( 20 ); \
      LCD_E( 0 ); \
      WAIT_US( 45 ); \
   }

#else

   #define LCD_DATA_4( x ){ \
      LATD &= 0x0F; \
      LATD |= (( x ) << 4 ); \
      WAIT_US( 20 ); \
      LCD_E( 1 ); \
      WAIT_US( 20 ); \
      LCD_E( 0 ); \
      WAIT_US( 45 ); \
   }

#endif
	
void lcd_data_4( unsigned char x ){
   LCD_DATA_4( x );
}

#define LCD_DATA_8( x ){ \
   LCD_DATA_4(( x ) >> 4 ); \
   LCD_DATA_4(( x ) & 0x0F ); \
}

int lcd_data_8( unsigned char x ){
   LCD_DATA_8( x );
}

unsigned char _lcd_lines, _lcd_columns;
unsigned char _lcd_x, _lcd_y;

void lcd_init( unsigned char x, unsigned char y ){
   PIN_DIRECTION( PORTA, 5, OUTPUT ); // RA5 output
   TRISD &= ( 0xF8 ^ 0xFF );  // RD3 .. RD7 outputs
   LCD_E( 0 );                // E idle
   LCD_RS( 0 );               // select 'command'
   wait_ms( 50 );             // wait for the HD44780 startup
	 
   // startup sequence as required by the HD44780 datasheet
   lcd_data_4( 0x03 );
   wait_ms( 5 );
   lcd_data_4( 0x03 );
   wait_ms( 5 );
   lcd_data_4( 0x03 );
   wait_ms( 5 );

   // select 4 bit mode
   lcd_data_4( 0x02 );       
   wait_ms( 5 );
         
   // functional initialisation
   lcd_command( 0x28 );      // 4 bit mode, 2 lines, 5x8 font
   wait_ms( 5 );
   lcd_command( 0x0C );      // display on, no cursor, no blink
   wait_ms( 5 );

   _lcd_lines = y;
   _lcd_columns = x;
}

void lcd_command( unsigned char x ){
   LCD_RS( 0 );
   LCD_DATA_8( x );  
}

void lcd_clear( void ){
   _lcd_x = 0;
   _lcd_y = 0;
   lcd_command( 0x01 );	
   wait_ms( 5 );
}

void lcd_line( unsigned char n ){
   _lcd_x = 0;
   _lcd_y = n;
   if( n == 0 ){ 
      lcd_command( 0x80 ); 
   } else { 
      lcd_command( 0xA8 ); 
   }
}

void lcd_xy( unsigned char x, unsigned char y ){
   _lcd_x = x;
   _lcd_y = y;
   if( _lcd_columns == 8 ){
      if( x < 8 ){
         lcd_command( 0x80 + x );
      } else {
         lcd_command( 0x80 + 0x40 + ( x - 8 ));
      }
   } else {
      lcd_command( 0x80 + 
        (( y > 0 ) ? 0x40 : 0x00 ) 
        + ( x ));
   }
}   

void lcd_data( unsigned char x ){
   LCD_RS( 1 );
   LCD_DATA_8( x );  
}

void lcd_char_print( char c ){
   if( _lcd_columns == 8 ){
      if( _lcd_x == 8 ){
         lcd_xy( 8, _lcd_y );
      }
   }
   if( 
      ( _lcd_y < _lcd_columns )
      && ( _lcd_x < _lcd_lines )
   ){
      lcd_data( c );
      _lcd_x++;
   }  
}

void lcd_string_print( const char *p ){
   while( *p != 0 ){
     if( *p == '\n' ){
        lcd_xy( 0, _lcd_y + 1 );
     } else if( *p == '\r' ){
        lcd_xy( 0, _lcd_y );
     } else if( *p == '\v' ){
        lcd_clear();
     } else {
        lcd_char_print( *p );
     }
     p++;
   }
}

#ifdef DB038_PIC16
   void lcd_const_string_print( const rom char *p ){
      while( *p != 0 ){
        lcd_char_print( *p );
        p++;
      }
   }
#endif

void lcd_custom( 
   unsigned char n,
   unsigned char d1,
   unsigned char d2,
   unsigned char d3,
   unsigned char d4,
   unsigned char d5,
   unsigned char d6,
   unsigned char d7, 
   unsigned char d8
){
   lcd_command( 0x40 | ( n << 3 ) );
   lcd_data( d1 );	
   lcd_data( d2 );	
   lcd_data( d3 );	
   lcd_data( d4 );	
   lcd_data( d5 );	
   lcd_data( d6 );	
   lcd_data( d7 );	
   lcd_data( d8 );	
}
